"""
https://github.github.com/gfm/#precedence
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_block_inline_precedence_012():
    """
    Test case 012:  Indicators of block structure always take precedence over indicators of inline structure.
    """

    # Arrange
    source_markdown = """- `one
- two`"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):`one:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):two`:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>`one</li>
<li>two`</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to allow for the details on the extension to be encapsulated.
"""


# pylint: disable=too-many-instance-attributes
class ExtensionDetails:
    """
    Class to allow for the details on the extension to be encapsulated.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        extension_id,
        extension_name,
        extension_description,
        extension_enabled_by_default,
        extension_version,
        extension_interface_version,
        extension_url=None,
        extension_configuration=None,
    ):
        (
            self.__extension_id,
            self.__extension_name,
            self.__extension_description,
            self.__extension_enabled_by_default,
            self.__extension_version,
            self.__extension_interface_version,
            self.__extension_url,
            self.__extension_configuration,
        ) = (
            extension_id,
            extension_name,
            extension_description,
            extension_enabled_by_default,
            extension_version,
            extension_interface_version,
            extension_url,
            extension_configuration,
        )

    # pylint: enable=too-many-arguments

    @property
    def extension_id(self):
        """
        Property to get the id of the extension.
        """
        return self.__extension_id

    @property
    def extension_name(self):
        """
        Property to get the name of the extension.
        """
        return self.__extension_name

    @property
    def extension_description(self):
        """
        Property to get the short description of the extension.
        """
        return self.__extension_description

    @property
    def extension_enabled_by_default(self):
        """
        Property to get whether the extension is enabled by default.
        """
        return self.__extension_enabled_by_default

    @property
    def extension_version(self):
        """
        Property to get the version of the extension.
        """
        return self.__extension_version

    @property
    def extension_interface_version(self):
        """
        Property to get the interface version of the extension.
        """
        return self.__extension_interface_version

    @property
    def extension_url(self):
        """
        Property to get the optional url for the extension.
        """
        return self.__extension_url

    @property
    def extension_configuration(self):
        """
        Property to get the optional configuration items for the extension.
        """
        return self.__extension_configuration


# pylint: enable=too-many-instance-attributes



"""
Module to implement a plugin that looks for headings that are not surrounded by
blank lines.
"""
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd022(Plugin):
    """
    Class to implement a plugin that looks for headings that are not surrounded by
    blank lines.
    """

    def __init__(self):
        super().__init__()
        self.__blank_line_count = None
        self.__did_above_line_count_match = None
        self.__start_heading_token = None
        self.__did_heading_end = None
        self.__lines_above = None
        self.__lines_below = None
        self.__start_heading_blank_line_count = None

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="blanks-around-headings,blanks-around-headers",
            plugin_id="MD022",
            plugin_enabled_by_default=True,
            plugin_description="Headings should be surrounded by blank lines.",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://github.com/jackdewinter/pymarkdown/blob/main/docs/rules/rule_md022.md",
            plugin_configuration="lines_above, lines_below",
        )

    @classmethod
    def __validate_configuration_value(cls, found_value):
        if found_value < 0:
            raise ValueError("Value must not be zero or a positive integer.")

    def initialize_from_config(self):
        """
        Event to allow the plugin to load configuration information.
        """
        self.__lines_above = self.plugin_configuration.get_integer_property(
            "lines_above",
            default_value=1,
            valid_value_fn=RuleMd022.__validate_configuration_value,
        )
        self.__lines_below = self.plugin_configuration.get_integer_property(
            "lines_below",
            default_value=1,
            valid_value_fn=RuleMd022.__validate_configuration_value,
        )

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        self.__blank_line_count = -1
        self.__did_above_line_count_match = False
        self.__start_heading_token = None
        self.__did_heading_end = False

    def completed_file(self, context):
        """
        Event that the file being currently scanned is now completed.
        """
        self.perform_close_check(context, None)

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """
        # print("START>>" + str(token).replace("\n", "\\n").replace("\t", "\\t"))
        # print(">>self.__blank_line_count>>" + str(self.__blank_line_count))
        self.perform_close_check(context, token)

        if token.is_blank_line:
            # print(">>token.is_blank_line>>")
            if (self.__blank_line_count is not None) and self.__blank_line_count >= 0:
                self.__blank_line_count += 1
        if token.is_setext_heading or token.is_atx_heading:
            # print(">>token.is_setext_heading or token.is_atx_heading>>")
            self.__did_above_line_count_match = bool(
                self.__blank_line_count == -1
                or self.__blank_line_count == self.__lines_above
            )
            self.__start_heading_token = token
            self.__start_heading_blank_line_count = self.__blank_line_count
            self.__did_heading_end = False
            # print("self.__did_above_line_count_match>>" + str(self.__did_above_line_count_match))
        elif token.is_thematic_break:
            self.__blank_line_count = 0
        elif token.is_end_token:
            # print(">>token.is_end_token>>")
            if not token.is_list_end and not token.is_block_quote_end:
                self.__blank_line_count = (
                    0
                    if token.is_leaf_end_token or token.is_container_end_token
                    else None
                )
            if token.is_atx_heading_end or token.is_setext_heading_end:
                self.__did_heading_end = True
        # print(">>self.__blank_line_count>>" + str(self.__blank_line_count))

    def perform_close_check(self, context, token):
        """
        Perform any state checks necessary upon closing the heading context.  Also
        called at the end of a document to make sure the implicit close of the
        document is handled properly.
        """

        if (
            (self.__start_heading_token and self.__did_heading_end)
        ) and self.__blank_line_count >= 0:
            if not token or (not token.is_blank_line and not token.is_block_quote_end):
                did_end_match = bool(self.__blank_line_count == self.__lines_below)
                self.report_any_match_failures(context, did_end_match)
                self.__start_heading_token = None

    def report_any_match_failures(self, context, did_end_match):
        """
        Take care of reporting any match failures.
        """

        if not self.__did_above_line_count_match:
            extra_info = (
                "Expected: "
                + str(self.__lines_above)
                + "; Actual: "
                + str(self.__start_heading_blank_line_count)
                + "; Above"
            )
            self.report_next_token_error(
                context,
                self.__start_heading_token,
                extra_error_information=extra_info,
                use_original_position=self.__start_heading_token.is_setext_heading,
            )
        if not did_end_match:
            extra_info = (
                "Expected: "
                + str(self.__lines_below)
                + "; Actual: "
                + str(self.__blank_line_count)
                + "; Below"
            )
            self.report_next_token_error(
                context,
                self.__start_heading_token,
                extra_error_information=extra_info,
                use_original_position=self.__start_heading_token.is_setext_heading,
            )



"""
Module to implement a plugin that looks for headings that do not start at the
beginning of the line.
"""
from pymarkdown.parser_helper import ParserHelper
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd023(Plugin):
    """
    Class to implement a plugin that looks for headings that do not start at the
    beginning of the line.
    """

    def __init__(self):
        super().__init__()
        self.__setext_start_token = None
        self.__any_leading_whitespace_detected = None
        self.__seen_first_line_of_setext = None

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="heading-start-left, header-start-left",
            plugin_id="MD023",
            plugin_enabled_by_default=True,
            plugin_description="Headings must start at the beginning of the line.",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://github.com/jackdewinter/pymarkdown/blob/main/docs/rules/rule_md023.md",
        )

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        self.__setext_start_token = None
        self.__any_leading_whitespace_detected = False
        self.__seen_first_line_of_setext = False

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """
        if token.is_atx_heading:
            if token.extracted_whitespace:
                self.report_next_token_error(context, token)
        elif token.is_setext_heading:
            self.__setext_start_token = token
            self.__any_leading_whitespace_detected = bool(token.extracted_whitespace)
            self.__seen_first_line_of_setext = False
        elif token.is_text:
            if (
                self.__setext_start_token
                and not self.__any_leading_whitespace_detected
                and token.end_whitespace
            ):
                split_end_whitespace = token.end_whitespace.split("\n")
                for next_split in split_end_whitespace:
                    if self.__seen_first_line_of_setext:
                        split_next_split = next_split.split(
                            ParserHelper.whitespace_split_character
                        )
                        if len(split_next_split) == 2 and split_next_split[0]:
                            self.__any_leading_whitespace_detected = True
                    else:
                        self.__seen_first_line_of_setext = True
        elif token.is_setext_heading_end:
            if token.extracted_whitespace:
                self.__any_leading_whitespace_detected = True

            if self.__any_leading_whitespace_detected:
                self.report_next_token_error(context, self.__setext_start_token)
            self.__setext_start_token = None



"""
Module to provide tests related to the MD026 rule.
"""
from test.markdown_scanner import MarkdownScanner

import pytest


@pytest.mark.rules
def test_md035_bad_configuration_style():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md035.style=$#1",
        "--strict-config",
        "scan",
        "test/resources/rules/md035/good_consistent_dash.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md035.style' must be of type 'str'."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_bad_configuration_style_leading_spaces():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md035.style= ---",
        "--strict-config",
        "scan",
        "test/resources/rules/md035/good_consistent_dash.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md035.style' is not valid: Allowable values cannot including leading or trailing spaces."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_good_configuration_style_consistent():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md035.style=consistent",
        "--strict-config",
        "scan",
        "test/resources/rules/md035/good_consistent_dash.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_bad_configuration_style_empty():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md035.style=",
        "--strict-config",
        "scan",
        "test/resources/rules/md035/good_consistent_dash.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md035.style' is not valid: Allowable values are: consistent, '---', '***', or any other horizontal rule text."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_bad_configuration_style_trailing_spaces():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md035.style=--- ",
        "--strict-config",
        "scan",
        "test/resources/rules/md035/good_consistent_dash.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md035.style' is not valid: Allowable values cannot including leading or trailing spaces."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_bad_configuration_style_bad_character():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md035.style=-=-=-",
        "--strict-config",
        "scan",
        "test/resources/rules/md035/good_consistent_dash.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md035.style' is not valid: Allowable values are: consistent, '---', '***', or any other horizontal rule text."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_good_consistent_dash():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md035/good_consistent_dash.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_bad_consistent_dash():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md035/bad_consistent_dash.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md035/bad_consistent_dash.md:5:1: "
        + "MD035: Horizontal rule style "
        + "[Expected: ---, Actual: - - -] (hr-style)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_bad_consistent_dash_with_leading_spaces():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md035/bad_consistent_dash_with_leading_spaces.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md035/bad_consistent_dash_with_leading_spaces.md:5:2: "
        + "MD035: Horizontal rule style "
        + "[Expected: ---, Actual: - - -] (hr-style)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_good_dash_marker():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md035.style=---",
        "--strict-config",
        "scan",
        "test/resources/rules/md035/good_consistent_dash.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_bad_dash_marker():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md035.style=---",
        "scan",
        "test/resources/rules/md035/bad_consistent_dash.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md035/bad_consistent_dash.md:5:1: "
        + "MD035: Horizontal rule style "
        + "[Expected: ---, Actual: - - -] (hr-style)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_good_consistent_asterisk():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md035/good_consistent_asterisk.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_bad_consistent_asterisk():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md035/bad_consistent_asterisk.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md035/bad_consistent_asterisk.md:5:1: "
        + "MD035: Horizontal rule style "
        + "[Expected: ***, Actual: * * *] (hr-style)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_good_asterisk_marker():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md035.style=* * *",
        "--strict-config",
        "scan",
        "test/resources/rules/md035/good_consistent_asterisk.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_bad_asterisk_marker():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md035.style=* * *",
        "scan",
        "test/resources/rules/md035/bad_consistent_asterisk.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md035/bad_consistent_asterisk.md:1:1: "
        + "MD035: Horizontal rule style "
        + "[Expected: * * *, Actual: ***] (hr-style)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_good_consistent_underscore():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md035/good_consistent_underscore.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_bad_consistent_underscore():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md035/bad_consistent_underscore.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md035/bad_consistent_underscore.md:5:1: "
        + "MD035: Horizontal rule style "
        + "[Expected: ___, Actual: ______] (hr-style)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_good_underscore_marker():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md035.style=______",
        "--strict-config",
        "scan",
        "test/resources/rules/md035/good_consistent_underscore.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md035_bad_underscore_marker():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md035.style=______",
        "scan",
        "test/resources/rules/md035/bad_consistent_underscore.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md035/bad_consistent_underscore.md:1:1: "
        + "MD035: Horizontal rule style "
        + "[Expected: ______, Actual: ___] (hr-style)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
Module to provide processing for the container blocks.
"""
import logging

from pymarkdown.block_quote_processor import BlockQuoteProcessor
from pymarkdown.extensions.pragma_token import PragmaExtension
from pymarkdown.html_helper import HtmlHelper
from pymarkdown.inline_markdown_token import TextMarkdownToken
from pymarkdown.leaf_block_processor import LeafBlockProcessor
from pymarkdown.link_reference_definition_helper import LinkReferenceDefinitionHelper
from pymarkdown.list_block_processor import ListBlockProcessor
from pymarkdown.parser_helper import ParserHelper, PositionMarker
from pymarkdown.parser_logger import ParserLogger

POGGER = ParserLogger(logging.getLogger(__name__))


class ContainerBlockProcessor:
    """
    Class to provide processing for the container blocks.
    """

    # pylint: disable=too-many-locals
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-statements
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-return-statements
    @staticmethod
    def parse_line_for_container_blocks(
        parser_state,
        position_marker,
        ignore_link_definition_start,
        parser_properties,
        container_start_bq_count,
        container_depth=0,
        foobar=None,
        init_bq=None,
    ):
        """
        Parse the line, taking care to handle any container blocks before deciding
        whether or not to pass the (remaining parts of the) line to the leaf block
        processor.

        Note: This is one of the more heavily traffic functions in the
        parser.  Debugging should be uncommented only if needed.
        """

        line_to_parse = position_marker.text_to_parse
        if container_depth == 0:
            parser_state.mark_start_information(position_marker)

        # POGGER.debug("Line:$:", line_to_parse)
        # POGGER.debug("Stack Depth:$:", parser_state.original_stack_depth)
        # POGGER.debug("Document Depth:$:", parser_state.original_document_depth)

        # Debug to be used for block quotes if needed.
        # POGGER.debug(
        #    "Last Block Quote:$:",
        #    parser_state.last_block_quote_stack_token,
        # )
        # POGGER.debug(
        #    "Last Block Quote:$:",
        #    parser_state.last_block_quote_markdown_token_index,
        # )
        # POGGER.debug(
        #    "Last Block Quote:$:", parser_state.copy_of_last_block_quote_markdown_token
        # )

        start_index, extracted_whitespace = ParserHelper.extract_whitespace(
            line_to_parse, 0
        )

        if ContainerBlockProcessor.__look_for_pragmas(
            position_marker,
            line_to_parse,
            container_depth,
            extracted_whitespace,
            parser_properties,
        ):
            return None, None, None

        (
            current_container_blocks,
            adj_ws,
            stack_bq_count,
            this_bq_count,
        ) = ContainerBlockProcessor.__calculate_for_container_blocks(
            parser_state,
            line_to_parse,
            extracted_whitespace,
            foobar,
            init_bq,
        )

        end_container_indices = ContainerIndices(-1, -1, -1)
        parser_state.nested_list_start = None
        (
            did_process,
            was_container_start,
            end_container_indices.block_index,
            this_bq_count,
            stack_bq_count,
            line_to_parse,
            start_index,
            leaf_tokens,
            container_level_tokens,
            removed_chars_at_start,
            did_blank,
            last_block_quote_index,
            text_removed_by_container,
            avoid_block_starts,
            requeue_line_info,
        ) = ContainerBlockProcessor.__get_block_start_index(
            position_marker,
            parser_state,
            extracted_whitespace,
            adj_ws,
            this_bq_count,
            stack_bq_count,
            start_index,
            container_start_bq_count,
        )
        if requeue_line_info:
            POGGER.debug(">>requeuing lines after looking for block start. returning.")
            return None, None, requeue_line_info

        if did_blank:
            POGGER.debug(">>already handled blank line. returning.")
            container_level_tokens.extend(leaf_tokens)
            return container_level_tokens, line_to_parse, None

        # POGGER.debug("this_bq_count>>$", this_bq_count)
        # POGGER.debug("stack_bq_count>>$", stack_bq_count)
        # POGGER.debug(">>avoid_block_starts>>$", avoid_block_starts)
        # POGGER.debug(">>did_process>>$", did_process)

        (
            did_process,
            was_container_start,
            end_container_indices.ulist_index,
            line_to_parse,
            removed_chars_at_start,
            requeue_line_info,
        ) = ContainerBlockProcessor.__get_list_start_index(
            position_marker,
            line_to_parse,
            start_index,
            True,
            parser_state,
            did_process,
            was_container_start,
            extracted_whitespace,
            adj_ws,
            stack_bq_count,
            this_bq_count,
            removed_chars_at_start,
            current_container_blocks,
            container_level_tokens,
        )
        if requeue_line_info:
            POGGER.debug(
                ">>requeuing lines after looking for ordered list start. returning."
            )
            return None, None, requeue_line_info
        # POGGER.debug("this_bq_count>>$", this_bq_count)
        # POGGER.debug("stack_bq_count>>$", stack_bq_count)

        (
            did_process,
            was_container_start,
            end_container_indices.olist_index,
            line_to_parse,
            removed_chars_at_start,
            requeue_line_info,
        ) = ContainerBlockProcessor.__get_list_start_index(
            position_marker,
            line_to_parse,
            start_index,
            False,
            parser_state,
            did_process,
            was_container_start,
            extracted_whitespace,
            adj_ws,
            stack_bq_count,
            this_bq_count,
            removed_chars_at_start,
            current_container_blocks,
            container_level_tokens,
        )
        if requeue_line_info:
            POGGER.debug(
                ">>requeuing lines after looking for unordered list start. returning."
            )
            return None, None, requeue_line_info
        # POGGER.debug("this_bq_count>>$", this_bq_count)
        # POGGER.debug("stack_bq_count>>$", stack_bq_count)

        # POGGER.debug("last_block_quote_index>>$", last_block_quote_index)
        # POGGER.debug("indices>>$", end_container_indices)
        # POGGER.debug("line_to_parse(after containers)>>$", line_to_parse)

        last_list_start_index = 0
        if end_container_indices.block_index != -1:
            assert last_block_quote_index in (
                end_container_indices.block_index - 1,
                end_container_indices.block_index,
            )
        elif end_container_indices.olist_index != -1:
            last_list_start_index = end_container_indices.olist_index
        elif end_container_indices.ulist_index != -1:
            last_list_start_index = end_container_indices.ulist_index

        if not parser_state.token_stack[-1].is_fenced_code_block:
            new_position_marker = PositionMarker(
                position_marker.line_number, start_index, line_to_parse
            )
            # POGGER.debug("this_bq_count>>$", this_bq_count)
            # POGGER.debug("stack_bq_count>>$", stack_bq_count)
            (
                line_to_parse,
                leaf_tokens,
                container_level_tokens,
            ) = ContainerBlockProcessor.__handle_nested_container_blocks(
                parser_state,
                container_depth,
                this_bq_count,
                stack_bq_count,
                new_position_marker,
                parser_properties,
                end_container_indices,
                leaf_tokens,
                container_level_tokens,
                was_container_start,
                avoid_block_starts,
                start_index,
                removed_chars_at_start,
                text_removed_by_container,
            )
            # POGGER.debug_with_visible_whitespace("text>>$>>", line_to_parse)
            # POGGER.debug("this_bq_count>>$", this_bq_count)
            # POGGER.debug("stack_bq_count>>$", stack_bq_count)

        # POGGER.debug("olist->container_level_tokens->$", container_level_tokens)
        # POGGER.debug("removed_chars_at_start>>>$", removed_chars_at_start)

        if container_depth:
            assert not leaf_tokens
            POGGER.debug(">>>>>>>>$<<<<<<<<<<", line_to_parse)
            return container_level_tokens, line_to_parse, None

        # POGGER.debug_with_visible_whitespace(
        #     ">>__process_list_in_progress>>$>>",
        #     line_to_parse,
        # )
        (
            did_process,
            line_to_parse,
            container_level_tokens,
            used_indent,
        ) = ContainerBlockProcessor.__process_list_in_progress(
            parser_state,
            did_process,
            line_to_parse,
            start_index,
            container_level_tokens,
            extracted_whitespace,
        )
        # POGGER.debug_with_visible_whitespace(
        #     ">>__process_list_in_progress>>$>>", line_to_parse
        # )
        # POGGER.debug("container_start_bq_count>>$", container_start_bq_count)
        # POGGER.debug("this_bq_count>>$", this_bq_count)
        # POGGER.debug("stack_bq_count>>$", stack_bq_count)
        ContainerBlockProcessor.__process_lazy_lines(
            parser_state,
            leaf_tokens,
            this_bq_count,
            stack_bq_count,
            line_to_parse,
            container_level_tokens,
            container_start_bq_count,
        )
        # POGGER.debug_with_visible_whitespace("text>>$>>", line_to_parse)
        # POGGER.debug("container_level_tokens>>$>>", container_level_tokens)

        # TODO refactor to make indent unnecessary?
        calculated_indent = len(parser_state.original_line_to_parse) - len(
            line_to_parse
        )
        # POGGER.debug(">>indent>>$", calculated_indent)

        force_it = False
        if (
            used_indent
            and parser_state.token_stack[-1].is_paragraph
            and parser_state.token_stack[-2].is_block_quote
        ):
            assert text_removed_by_container is None
            text_removed_by_container, force_it = used_indent, True

        newer_position_marker = PositionMarker(
            position_marker.line_number,
            start_index,
            line_to_parse,
            index_indent=calculated_indent,
        )
        parser_state.mark_for_leaf_processing(container_level_tokens)
        leaf_tokens, requeue_line_info = ContainerBlockProcessor.__process_leaf_tokens(
            parser_state,
            leaf_tokens,
            newer_position_marker,
            this_bq_count,
            removed_chars_at_start,
            ignore_link_definition_start,
            last_block_quote_index,
            last_list_start_index,
            text_removed_by_container,
            force_it,
        )
        parser_state.clear_after_leaf_processing()

        container_level_tokens.extend(leaf_tokens)
        return container_level_tokens, line_to_parse, requeue_line_info
        # pylint: enable=too-many-locals
        # pylint: enable=too-many-arguments
        # pylint: enable=too-many-statements
        # pylint: enable=too-many-branches
        # pylint: enable=too-many-return-statements

    @staticmethod
    # pylint: disable=too-many-locals, too-many-arguments
    def __get_block_start_index(
        position_marker,
        parser_state,
        extracted_whitespace,
        adj_ws,
        this_bq_count,
        stack_bq_count,
        start_index,
        container_start_bq_count,
    ):
        new_position_marker = PositionMarker(
            position_marker.line_number, start_index, position_marker.text_to_parse
        )
        (
            did_process,
            was_container_start,
            block_index,
            this_bq_count,
            stack_bq_count,
            line_to_parse,
            start_index,
            leaf_tokens,
            container_level_tokens,
            removed_chars_at_start,
            did_blank,
            last_block_quote_index,
            text_removed_by_container,
            avoid_block_starts,
            requeue_line_info,
        ) = BlockQuoteProcessor.handle_block_quote_block(
            parser_state,
            new_position_marker,
            extracted_whitespace,
            adj_ws,
            this_bq_count,
            stack_bq_count,
            container_start_bq_count,
        )
        # POGGER.debug("container_start_bq_count>>:$", container_start_bq_count)
        # POGGER.debug("this_bq_count>>:$", this_bq_count)
        # POGGER.debug("stack_bq_count>>$", stack_bq_count)
        # POGGER.debug("did_process>>$", did_process)
        POGGER.debug("text>>:$:>>", line_to_parse)
        POGGER.debug(">>container_level_tokens>>$", container_level_tokens)
        return (
            did_process,
            was_container_start,
            block_index,
            this_bq_count,
            stack_bq_count,
            line_to_parse,
            start_index,
            leaf_tokens,
            container_level_tokens,
            removed_chars_at_start,
            did_blank,
            last_block_quote_index,
            text_removed_by_container,
            avoid_block_starts,
            requeue_line_info,
        )

    # pylint: enable=too-many-locals, too-many-arguments

    # pylint: disable=too-many-locals, too-many-arguments
    @staticmethod
    def __get_list_start_index(
        position_marker,
        line_to_parse,
        start_index,
        is_ulist,
        parser_state,
        did_process,
        was_container_start,
        extracted_whitespace,
        adj_ws,
        stack_bq_count,
        this_bq_count,
        removed_chars_at_start,
        current_container_blocks,
        container_level_tokens,
    ):
        """
        Note: This is one of the more heavily traffic functions in the
        parser.  Debugging should be uncommented only if needed.
        """
        # TODO refactor so it doesn't need this!
        new_position_marker = PositionMarker(
            position_marker.line_number, start_index, line_to_parse
        )

        POGGER.debug(
            "pre-list>>#$#$#$#",
            position_marker.index_number,
            position_marker.index_indent,
            position_marker.text_to_parse,
        )
        POGGER.debug(
            "pre-list>>#$#$#$#",
            new_position_marker.index_number,
            new_position_marker.index_indent,
            new_position_marker.text_to_parse,
        )
        (
            did_process,
            was_container_start,
            new_list_index,
            line_to_parse,
            resultant_tokens,
            removed_chars_at_start,
            requeue_line_info,
        ) = ListBlockProcessor.handle_list_block(
            is_ulist,
            parser_state,
            did_process,
            was_container_start,
            new_position_marker,
            extracted_whitespace,
            adj_ws,
            stack_bq_count,
            this_bq_count,
            removed_chars_at_start,
            current_container_blocks,
        )
        if requeue_line_info:
            return (
                None,
                None,
                None,
                None,
                None,
                requeue_line_info,
            )
        container_level_tokens.extend(resultant_tokens)
        POGGER.debug(
            "post-ulist>>#$#$#$#",
            position_marker.index_number,
            position_marker.index_indent,
            position_marker.text_to_parse,
        )
        POGGER.debug(
            "post-ulist>>#$#$#$#",
            new_position_marker.index_number,
            new_position_marker.index_indent,
            new_position_marker.text_to_parse,
        )
        POGGER.debug("text>>$>>", line_to_parse)

        return (
            did_process,
            was_container_start,
            new_list_index,
            line_to_parse,
            removed_chars_at_start,
            None,
        )

    # pylint: enable=too-many-locals, too-many-arguments

    @staticmethod
    def __calculate_for_container_blocks(
        parser_state,
        line_to_parse,
        extracted_whitespace,
        foobar,
        init_bq,
    ):
        """
        Perform some calculations that will be needed for parsing the container blocks.
        """
        this_bq_count = 0 if init_bq is None else init_bq

        current_container_blocks = [
            ind for ind in parser_state.token_stack if ind.is_list
        ]

        adj_ws = ContainerBlockProcessor.__calculate_adjusted_whitespace(
            parser_state,
            current_container_blocks,
            line_to_parse,
            extracted_whitespace,
            foobar=foobar,
        )

        return (
            current_container_blocks,
            adj_ws,
            parser_state.count_of_block_quotes_on_stack(),
            this_bq_count,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __calculate_adjusted_whitespace(
        parser_state,
        current_container_blocks,
        line_to_parse,
        extracted_whitespace,
        foobar=None,
        previous_ws_len=0,
    ):
        """
        Based on the last container on the stack, determine what the adjusted whitespace is.
        """

        adj_ws, stack_index = (
            extracted_whitespace,
            parser_state.find_last_list_block_on_stack(),
        )
        if stack_index <= 0:
            POGGER.debug("PLFCB>>No Started lists")
            assert not current_container_blocks
            if foobar is None:
                POGGER.debug("PLFCB>>No Started Block Quote")
            else:
                POGGER.debug("PLFCB>>Started Block Quote")
                adj_ws = extracted_whitespace[foobar:]
        else:
            assert current_container_blocks
            POGGER.debug(
                "PLFCB>>Started list-last stack>>$",
                parser_state.token_stack,
            )
            POGGER.debug(
                "PLFCB>>Started list-last stack>>$",
                parser_state.token_stack[stack_index],
            )
            token_index = len(parser_state.token_document) - 1

            while token_index >= 0 and not (
                parser_state.token_document[token_index].is_any_list_token
            ):
                token_index -= 1
            POGGER.debug(
                "PLFCB>>Started list-last token>>$",
                parser_state.token_document[token_index],
            )
            assert token_index >= 0

            old_start_index = parser_state.token_document[token_index].indent_level

            ws_len = (
                ParserHelper.calculate_length(extracted_whitespace) + previous_ws_len
            )
            POGGER.debug("old_start_index>>$>>ws_len>>$", old_start_index, ws_len)
            if ws_len >= old_start_index:
                POGGER.debug("RELINE:$:", line_to_parse)
                adj_ws = extracted_whitespace[old_start_index:]
        return adj_ws

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals, too-many-statements
    @staticmethod
    def __handle_nested_container_blocks(
        parser_state,
        container_depth,
        this_bq_count,
        stack_bq_count,
        position_marker,
        parser_properties,
        end_container_indices,
        leaf_tokens,
        container_level_tokens,
        was_container_start,
        avoid_block_starts,
        start_index,
        removed_chars_at_start,
        text_removed_by_container,
    ):
        """
        Handle the processing of nested container blocks, as they can contain
        themselves and get somewhat messy.
        """
        adjusted_text_to_parse = position_marker.text_to_parse

        POGGER.debug("adjusted_text_to_parse>$<", adjusted_text_to_parse)
        POGGER.debug("index_number>$<", position_marker.index_number)
        POGGER.debug("index_indent>$<", position_marker.index_indent)
        POGGER.debug("start_index>$<", start_index)
        POGGER.debug("parser_state.nested_list_start>$", parser_state.nested_list_start)

        if was_container_start and position_marker.text_to_parse:
            assert container_depth < 10
            POGGER.debug(
                "__handle_nested_container_blocks>stack>>:$:<<",
                position_marker.text_to_parse,
            )
            POGGER.debug(
                "__handle_nested_container_blocks>end_container_indices>>:$:<<",
                end_container_indices,
            )
            nested_container_starts = (
                ContainerBlockProcessor.__get_nested_container_starts(
                    parser_state,
                    position_marker.text_to_parse,
                    end_container_indices,
                    avoid_block_starts,
                    start_index,
                    removed_chars_at_start,
                    text_removed_by_container,
                )
            )
            POGGER.debug(
                "__handle_nested_container_blocks>nested_container_starts>>:$:<<",
                nested_container_starts,
            )
            POGGER.debug(
                "__handle_nested_container_blocks>end_container_indices>>:$:<<",
                end_container_indices,
            )

            POGGER.debug(
                "check next container_start>stack>>$", parser_state.token_stack
            )
            POGGER.debug("check next container_start>leaf_tokens>>$", leaf_tokens)
            POGGER.debug(
                "check next container_start>container_level_tokens>>$",
                container_level_tokens,
            )

            adj_line_to_parse = position_marker.text_to_parse

            POGGER.debug("check next container_start>pre>>$<<", adj_line_to_parse)
            active_container_index = max(
                end_container_indices.ulist_index,
                end_container_indices.olist_index,
                end_container_indices.block_index,
            )
            POGGER.debug(
                "check next container_start>max>>$>>bq>>$",
                active_container_index,
                end_container_indices.block_index,
            )
            delta = 0
            already_adjusted = False
            if (
                end_container_indices.block_index != -1
                and not nested_container_starts.ulist_index
                and not nested_container_starts.olist_index
            ):
                assert active_container_index == end_container_indices.block_index
                POGGER.debug(
                    "parser_state.nested_list_start>>$<<",
                    parser_state.nested_list_start,
                )
                POGGER.debug("adj_line_to_parse>>$<<", adj_line_to_parse)
                POGGER.debug(
                    "parser_state.token_document>>$<<", parser_state.token_document
                )
                if parser_state.nested_list_start and adj_line_to_parse.strip():
                    start_index, _ = ParserHelper.extract_whitespace(
                        adj_line_to_parse, 0
                    )
                    POGGER.debug(
                        "end_container_indices.block_index>>$<<",
                        end_container_indices.block_index,
                    )
                    POGGER.debug("start_index>>$<<", start_index)

                    indent_level = parser_state.nested_list_start.indent_level
                    POGGER.debug(
                        "indent_level>>$<<",
                        indent_level,
                    )

                    POGGER.debug(
                        "parser_state.nested_list_start.matching_markdown_token>>$<<",
                        parser_state.nested_list_start.matching_markdown_token,
                    )
                    list_start_token_index = parser_state.token_document.index(
                        parser_state.nested_list_start.matching_markdown_token
                    )
                    POGGER.debug(
                        "list_start_token_index>>$<<",
                        list_start_token_index,
                    )
                    token_after_list_start = parser_state.token_document[
                        list_start_token_index + 1
                    ]
                    POGGER.debug(
                        "token_after_list_start>>$<<",
                        token_after_list_start,
                    )
                    assert (
                        parser_state.nested_list_start.matching_markdown_token.line_number
                        == token_after_list_start.line_number
                    )
                    column_number_delta = (
                        token_after_list_start.column_number
                        - parser_state.nested_list_start.matching_markdown_token.column_number
                    )
                    POGGER.debug(
                        "column_number_delta>>$<<",
                        column_number_delta,
                    )
                    adjusted_indent_level = (
                        column_number_delta + end_container_indices.block_index
                    )
                    POGGER.debug(
                        "adjusted_indent_level>>$<<  indent_level>$",
                        adjusted_indent_level,
                        indent_level,
                    )
                    indent_was_adjusted = indent_level != adjusted_indent_level
                    if indent_level > adjusted_indent_level:
                        delta = indent_level - adjusted_indent_level
                    indent_level = (
                        column_number_delta + end_container_indices.block_index
                    )

                    if (
                        parser_state.token_document[-1].is_blank_line
                        and (end_container_indices.block_index + start_index)
                        < indent_level
                    ):
                        POGGER.debug("\n\nBOOM\n\n")

                        y_tokens = []
                        while parser_state.token_document[-1].is_blank_line:
                            y_tokens.append(parser_state.token_document[-1])
                            del parser_state.token_document[-1]

                        x_tokens, _ = parser_state.close_open_blocks_fn(
                            parser_state,
                            include_lists=True,
                        )
                        # assert False
                        parser_state.token_document.extend(x_tokens)
                        parser_state.token_document.extend(y_tokens)
                    elif delta:
                        POGGER.debug("adj_line_to_parse>>$<<", adj_line_to_parse)
                        adj_line_to_parse = adj_line_to_parse[delta:]
                        POGGER.debug("adj_line_to_parse>>$<<", adj_line_to_parse)
                    elif (
                        not nested_container_starts.block_index
                        and adj_line_to_parse
                        and adj_line_to_parse[0] == " "
                        and indent_was_adjusted
                        and parser_state.nested_list_start
                    ):

                        POGGER.debug("adj_line_to_parse>:$:<", adj_line_to_parse)
                        POGGER.debug(
                            "parser_state.nested_list_start>:$:<",
                            parser_state.nested_list_start.matching_markdown_token,
                        )
                        POGGER.debug("BOOM")
                        assert adj_line_to_parse.startswith(
                            parser_state.nested_list_start.matching_markdown_token.extracted_whitespace
                        )
                        adj_line_to_parse = adj_line_to_parse[
                            len(
                                parser_state.nested_list_start.matching_markdown_token.extracted_whitespace
                            ) :
                        ]
                        already_adjusted = True

            POGGER.debug(
                "check next container_start>mid>>stack_bq_count>>$<<this_bq_count<<$",
                stack_bq_count,
                this_bq_count,
            )
            if delta or already_adjusted:
                POGGER.debug(
                    "check next container_start>already adjusted<<$<<",
                    adj_line_to_parse,
                )
                adjusted_text_to_parse = adj_line_to_parse
            else:
                POGGER.debug("check next container_start>post<<$<<", adj_line_to_parse)
                adj_line_to_parse = f"{ParserHelper.repeat_string(ParserHelper.space_character, active_container_index)}{adj_line_to_parse}"
                POGGER.debug("check next container_start>post<<$<<", adj_line_to_parse)

            POGGER.debug("leaf_tokens>>$", leaf_tokens)
            assert not leaf_tokens
            if container_level_tokens:
                parser_state.token_document.extend(container_level_tokens)
                container_level_tokens = []

            POGGER.debug(
                "check next container_start>stack>>$", parser_state.token_stack
            )
            POGGER.debug(
                "check next container_start>tokenized_document>>$",
                parser_state.token_document,
            )

            if (
                nested_container_starts.ulist_index
                or nested_container_starts.olist_index
                or nested_container_starts.block_index
            ):
                POGGER.debug(
                    "check next container_start>nested_container",
                )
                adjusted_text_to_parse = (
                    ContainerBlockProcessor.__look_for_container_blocks(
                        parser_state,
                        adj_line_to_parse,
                        end_container_indices.block_index,
                        container_depth,
                        this_bq_count,
                        position_marker,
                        parser_properties,
                    )
                )
            parser_state.set_no_para_start_if_empty()
        return (
            adjusted_text_to_parse,
            leaf_tokens,
            container_level_tokens,
        )

    # pylint: enable=too-many-arguments, too-many-locals, too-many-statements

    # pylint: disable=too-many-arguments, too-many-locals, too-many-statements
    @staticmethod
    def __get_nested_container_starts(
        parser_state,
        line_to_parse,
        end_container_indices,
        avoid_block_starts,
        start_index,
        removed_chars_at_start,
        text_removed_by_container,
    ):

        POGGER.debug("check next container_start>")
        POGGER.debug("check next container_start>start_index>$", start_index)
        POGGER.debug(
            "check next container_start>removed_chars_at_start:$:",
            removed_chars_at_start,
        )
        POGGER.debug(
            "check next container_start>text_removed_by_container:$:",
            text_removed_by_container,
        )
        POGGER.debug("check next container_start>stack>>$", parser_state.token_stack)

        _, ex_ws_test = ParserHelper.extract_whitespace(line_to_parse, 0)

        whitespace_scan_start_index = 0
        for token_stack_index in parser_state.token_stack:
            if token_stack_index.is_block_quote:
                # if text_removed_by_container:
                #     if text_removed_by_container.startswith("> "):
                #         text_removed_by_container = text_removed_by_container[2:]
                #     elif text_removed_by_container.startswith(">"):
                #         text_removed_by_container = text_removed_by_container[1:]
                #     else:
                #         POGGER.info("check next container_start> out of block quote data")
                pass
            elif token_stack_index.is_list:
                if token_stack_index.ws_before_marker <= len(ex_ws_test):
                    whitespace_scan_start_index = token_stack_index.ws_before_marker

        after_ws_index, ex_whitespace = ParserHelper.extract_whitespace(
            line_to_parse, whitespace_scan_start_index
        )
        if not ex_whitespace:
            ex_whitespace = ""
            after_ws_index = whitespace_scan_start_index

        nested_ulist_start, _, _, _ = ListBlockProcessor.is_ulist_start(
            parser_state, line_to_parse, after_ws_index, ex_whitespace, False
        )
        nested_olist_start, _, _, _ = ListBlockProcessor.is_olist_start(
            parser_state, line_to_parse, after_ws_index, ex_whitespace, False
        )
        nested_block_start = (
            False
            if avoid_block_starts
            else BlockQuoteProcessor.is_block_quote_start(
                line_to_parse, after_ws_index, ex_whitespace
            )
        )
        POGGER.debug(
            "check next container_start>ulist>$>index>$",
            nested_ulist_start,
            end_container_indices.ulist_index,
        )
        POGGER.debug(
            "check next container_start>olist>$>index>$",
            nested_olist_start,
            end_container_indices.olist_index,
        )
        POGGER.debug(
            "check next container_start>bquote>$>index>$",
            nested_block_start,
            end_container_indices.block_index,
        )
        return ContainerIndices(
            nested_ulist_start, nested_olist_start, nested_block_start
        )

    # pylint: enable=too-many-arguments, too-many-locals, too-many-statements

    # pylint: disable=too-many-arguments
    @staticmethod
    def __look_for_container_blocks(
        parser_state,
        adj_line_to_parse,
        end_of_bquote_start_index,
        container_depth,
        this_bq_count,
        position_marker,
        parser_properties,
    ):
        """
        Look for container blocks that we can use.
        """
        POGGER.debug("check next container_start>recursing")
        POGGER.debug("check next container_start>>$\n", adj_line_to_parse)
        POGGER.debug("this_bq_count>$", this_bq_count)
        container_start_bq_count = this_bq_count

        adj_block = (
            None if end_of_bquote_start_index == -1 else end_of_bquote_start_index
        )

        position_marker = PositionMarker(
            position_marker.line_number, -1, adj_line_to_parse
        )
        (
            produced_inner_tokens,
            line_to_parse,
            requeue_line_info,
        ) = ContainerBlockProcessor.parse_line_for_container_blocks(
            parser_state,
            position_marker,
            False,
            parser_properties,
            container_start_bq_count,
            container_depth=container_depth + 1,
            foobar=adj_block,
            init_bq=this_bq_count,
        )
        assert not requeue_line_info or not requeue_line_info.lines_to_requeue
        # TODO will need to deal with force_ignore_first_as_lrd

        POGGER.debug("\ncheck next container_start>recursed")
        POGGER.debug("check next container_start>stack>>$", parser_state.token_stack)
        POGGER.debug(
            "check next container_start>tokenized_document>>$",
            parser_state.token_document,
        )
        POGGER.debug("check next container_start>line_parse>>$", line_to_parse)

        parser_state.token_document.extend(produced_inner_tokens)
        return line_to_parse

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __process_list_in_progress(
        parser_state,
        did_process,
        line_to_parse,
        start_index,
        container_level_tokens,
        extracted_whitespace,
    ):
        used_indent = None
        if not did_process:
            is_list_in_process, ind = LeafBlockProcessor.check_for_list_in_process(
                parser_state
            )
            if is_list_in_process:
                assert not container_level_tokens
                POGGER.debug("clt>>list-in-progress")
                POGGER.debug("clt>>line_to_parse>>:$:>>", line_to_parse)
                (
                    container_level_tokens,
                    line_to_parse,
                    used_indent,
                ) = ListBlockProcessor.list_in_process(
                    parser_state,
                    line_to_parse,
                    start_index,
                    extracted_whitespace,
                    ind,
                )
                POGGER.debug("clt>>line_to_parse>>:$:>>", line_to_parse)
                POGGER.debug("clt>>used_indent>>:$:>>", used_indent)
                did_process = True

        return did_process, line_to_parse, container_level_tokens, used_indent

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __process_lazy_lines(
        parser_state,
        leaf_tokens,
        this_bq_count,
        stack_bq_count,
        line_to_parse,
        container_level_tokens,
        container_start_bq_count,
    ):

        POGGER.debug("LINE-lazy>$", line_to_parse)
        assert not leaf_tokens
        POGGER.debug("clt>>lazy-check")

        POGGER.debug("__process_lazy_lines>>ltp>$", line_to_parse)
        after_ws_index, ex_whitespace = ParserHelper.extract_whitespace(
            line_to_parse, 0
        )
        remaining_line = line_to_parse[after_ws_index:]
        POGGER.debug("container_start_bq_count>>:$", container_start_bq_count)
        POGGER.debug("__process_lazy_lines>>this_bq_count>$<", this_bq_count)
        POGGER.debug("__process_lazy_lines>>stack_bq_count>$<", stack_bq_count)
        POGGER.debug("__process_lazy_lines>>mod->ltp>$<", remaining_line)
        POGGER.debug("__process_lazy_lines>>mod->ews>$<", ex_whitespace)

        lazy_tokens = BlockQuoteProcessor.check_for_lazy_handling(
            parser_state,
            this_bq_count,
            stack_bq_count,
            remaining_line,
            ex_whitespace,
        )
        if lazy_tokens:
            POGGER.debug("clt>>lazy-found")
            container_level_tokens.extend(lazy_tokens)

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __process_leaf_tokens(
        parser_state,
        leaf_tokens,
        xposition_marker,
        this_bq_count,
        removed_chars_at_start,
        ignore_link_definition_start,
        last_block_quote_index,
        last_list_start_index,
        text_removed_by_container,
        force_it,
    ):
        assert not leaf_tokens
        POGGER.debug("parsing leaf>>")
        position_marker = PositionMarker(
            xposition_marker.line_number,
            0,
            xposition_marker.text_to_parse,
            index_indent=xposition_marker.index_indent,
        )
        (
            leaf_tokens,
            requeue_line_info,
        ) = ContainerBlockProcessor.__parse_line_for_leaf_blocks(
            parser_state,
            position_marker,
            this_bq_count,
            removed_chars_at_start,
            ignore_link_definition_start,
            last_block_quote_index,
            last_list_start_index,
            text_removed_by_container,
            force_it,
        )
        POGGER.debug("parsed leaf>>$", leaf_tokens)
        return leaf_tokens, requeue_line_info

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __close_indented_block_if_indent_not_there(parser_state, extracted_whitespace):

        POGGER.debug(
            "__close_indented_block_if_indent_not_there>>$>",
            parser_state.token_stack[-1],
        )
        POGGER.debug(
            "__close_indented_block_if_indent_not_there>>$>", extracted_whitespace
        )
        pre_tokens = []
        if parser_state.token_stack[
            -1
        ].is_indented_code_block and ParserHelper.is_length_less_than_or_equal_to(
            extracted_whitespace, 3
        ):
            pre_tokens.append(
                parser_state.token_stack[
                    -1
                ].generate_close_markdown_token_from_stack_token()
            )
            del parser_state.token_stack[-1]

            extracted_blank_line_tokens = (
                ContainerBlockProcessor.extract_markdown_tokens_back_to_blank_line(
                    parser_state, False
                )
            )
            extracted_blank_line_tokens.reverse()
            pre_tokens.extend(extracted_blank_line_tokens)
        POGGER.debug(
            "__close_indented_block_if_indent_not_there>>pre_tokens>$>", pre_tokens
        )
        return pre_tokens

    @staticmethod
    def __handle_fenced_code_block(
        parser_state,
        outer_processed,
        position_marker,
        extracted_whitespace,
        new_tokens,
    ):
        """
        Take care of the processing for fenced code blocks.
        """
        if not parser_state.token_stack[-1].was_link_definition_started:
            (
                fenced_tokens,
                extracted_whitespace,
            ) = LeafBlockProcessor.parse_fenced_code_block(
                parser_state,
                position_marker,
                extracted_whitespace,
            )
            outer_processed = False
            if fenced_tokens:
                new_tokens.extend(fenced_tokens)
                outer_processed = True
            elif parser_state.token_stack[-1].is_fenced_code_block:
                new_tokens.append(
                    TextMarkdownToken(
                        position_marker.text_to_parse[position_marker.index_number :],
                        extracted_whitespace,
                        position_marker=position_marker,
                    )
                )
                outer_processed = True
        return outer_processed

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_html_block(
        parser_state,
        outer_processed,
        position_marker,
        extracted_whitespace,
        new_tokens,
    ):
        """
        Take care of the processing for html blocks.
        """

        POGGER.debug(">>position_marker>>ttp>>$>>", position_marker.text_to_parse)
        POGGER.debug(">>position_marker>>in>>$>>", position_marker.index_number)
        POGGER.debug(">>position_marker>>ln>>$>>", position_marker.line_number)
        if not outer_processed and not parser_state.token_stack[-1].is_html_block:
            POGGER.debug(">>html started?>>")
            old_top_of_stack = parser_state.token_stack[-1]
            html_tokens = HtmlHelper.parse_html_block(
                parser_state,
                position_marker,
                extracted_whitespace,
            )
            if html_tokens:
                POGGER.debug(">>html started>>")
                LeafBlockProcessor.correct_for_leaf_block_start_in_list(
                    parser_state,
                    position_marker.index_indent,
                    old_top_of_stack,
                    html_tokens,
                )
            else:
                POGGER.debug(">>html not started>>")
            new_tokens.extend(html_tokens)
        if parser_state.token_stack[-1].is_html_block:
            POGGER.debug(">>html continued>>")
            html_tokens = HtmlHelper.check_normal_html_block_end(
                parser_state,
                position_marker.text_to_parse,
                position_marker.index_number,
                extracted_whitespace,
                position_marker,
            )
            assert html_tokens
            new_tokens.extend(html_tokens)
            outer_processed = True

        return outer_processed

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_link_reference_definition(
        parser_state,
        outer_processed,
        position_marker,
        extracted_whitespace,
        remaining_line_to_parse,
        ignore_link_definition_start,
        pre_tokens,
    ):
        """
        Take care of the processing for link reference definitions.
        """
        requeue_line_info, new_tokens = None, []

        POGGER.debug(
            "handle_link_reference_definition>>pre_tokens>>$<<",
            pre_tokens,
        )

        if not outer_processed and not ignore_link_definition_start:
            POGGER.debug(
                "plflb-process_link_reference_definition>>outer_processed>>$",
                position_marker.text_to_parse[position_marker.index_number :],
            )
            (
                outer_processed,
                _,  # did_complete_lrd,
                _,  # did_pause_lrd,
                requeue_line_info,
                new_tokens,
            ) = LinkReferenceDefinitionHelper.process_link_reference_definition(
                parser_state,
                position_marker,
                remaining_line_to_parse,
                extracted_whitespace,
                parser_state.original_line_to_parse,
                parser_state.original_stack_depth,
                parser_state.original_document_depth,
            )
            if requeue_line_info and requeue_line_info.lines_to_requeue:
                outer_processed = True
                POGGER.debug(
                    "plflb-process_link_reference_definition>>outer_processed>>$<lines_to_requeue<$<$",
                    outer_processed,
                    requeue_line_info.lines_to_requeue,
                    len(requeue_line_info.lines_to_requeue),
                )
            else:
                POGGER.debug(
                    "plflb-process_link_reference_definition>>outer_processed>>$<lines_to_requeue<(None)",
                    outer_processed,
                )

        POGGER.debug("handle_link_reference_definition>>pre_tokens>>$<<", pre_tokens)
        pre_tokens.extend(new_tokens)
        POGGER.debug("handle_link_reference_definition>>pre_tokens>>$<<", pre_tokens)
        return outer_processed, requeue_line_info

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments,too-many-locals
    @staticmethod
    def __parse_line_for_leaf_blocks(
        parser_state,
        xposition_marker,
        this_bq_count,
        removed_chars_at_start,
        ignore_link_definition_start,
        last_block_quote_index,
        last_list_start_index,
        text_removed_by_container,
        force_it,
    ):
        """
        Parse the contents of a line for a leaf block.

        Note: This is one of the more heavily traffic functions in the
        parser.  Debugging should be uncommented only if needed.
        """
        POGGER.debug("Leaf Line:$:", xposition_marker.text_to_parse)
        new_tokens = []

        # TODO rename to avoid collision with parameter
        remaining_line_to_parse = xposition_marker.text_to_parse[
            xposition_marker.index_number :
        ]
        (new_index_number, extracted_whitespace,) = ParserHelper.extract_whitespace(
            xposition_marker.text_to_parse, xposition_marker.index_number
        )
        position_marker = PositionMarker(
            xposition_marker.line_number,
            new_index_number,
            xposition_marker.text_to_parse,
            index_indent=xposition_marker.index_indent,
        )

        pre_tokens = ContainerBlockProcessor.__close_indented_block_if_indent_not_there(
            parser_state, extracted_whitespace
        )

        outer_processed = ContainerBlockProcessor.__handle_fenced_code_block(
            parser_state,
            False,
            position_marker,
            extracted_whitespace,
            new_tokens,
        )

        (
            outer_processed,
            requeue_line_info,
        ) = ContainerBlockProcessor.__handle_link_reference_definition(
            parser_state,
            outer_processed,
            position_marker,
            extracted_whitespace,
            remaining_line_to_parse,
            ignore_link_definition_start,
            pre_tokens,
        )

        outer_processed = ContainerBlockProcessor.__handle_html_block(
            parser_state,
            outer_processed,
            position_marker,
            extracted_whitespace,
            new_tokens,
        )

        if not outer_processed:
            assert not new_tokens
            new_tokens = LeafBlockProcessor.parse_atx_headings(
                parser_state, position_marker, extracted_whitespace
            )
            if not new_tokens:
                new_tokens = LeafBlockProcessor.parse_indented_code_block(
                    parser_state,
                    position_marker,
                    extracted_whitespace,
                    removed_chars_at_start,
                    last_block_quote_index,
                    last_list_start_index,
                )
            if not new_tokens:
                stack_bq_count = parser_state.count_of_block_quotes_on_stack()
                new_tokens = LeafBlockProcessor.parse_setext_headings(
                    parser_state,
                    position_marker,
                    extracted_whitespace,
                    this_bq_count,
                    stack_bq_count,
                )
            if not new_tokens:
                stack_bq_count = parser_state.count_of_block_quotes_on_stack()
                new_tokens = LeafBlockProcessor.parse_thematic_break(
                    parser_state,
                    position_marker,
                    extracted_whitespace,
                    this_bq_count,
                    stack_bq_count,
                )
            if not new_tokens:
                stack_bq_count = parser_state.count_of_block_quotes_on_stack()
                new_tokens = LeafBlockProcessor.parse_paragraph(
                    parser_state,
                    position_marker,
                    extracted_whitespace,
                    this_bq_count,
                    stack_bq_count,
                    text_removed_by_container,
                    force_it,
                )

        POGGER.debug(">>leaf--adding>>$", new_tokens)
        pre_tokens.extend(new_tokens)
        POGGER.debug(">>leaf--added>>$", pre_tokens)
        return pre_tokens, requeue_line_info

    @staticmethod
    def extract_markdown_tokens_back_to_blank_line(parser_state, was_forced):
        """
        Extract tokens going back to the last blank line token.
        """

        pre_tokens = []
        while parser_state.token_document[-1].is_blank_line:
            last_element = parser_state.token_document[-1]
            if was_forced:
                pre_tokens.insert(0, last_element)
            else:
                pre_tokens.append(last_element)
            del parser_state.token_document[-1]
        return pre_tokens

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __look_for_pragmas(
        position_marker,
        line_to_parse,
        container_depth,
        extracted_whitespace,
        parser_properties,
    ):

        found_pragmas = False
        if parser_properties.is_pragmas_enabled:
            found_pragmas = PragmaExtension.look_for_pragmas(
                position_marker,
                line_to_parse,
                container_depth,
                extracted_whitespace,
                parser_properties,
            )
        return found_pragmas

    # pylint: enable=too-many-arguments


# pylint: disable=too-few-public-methods
# pylint: disable=too-many-lines
class ContainerIndices:
    """
    Class to provide for encapsulation on a group of container indices.
    """

    def __init__(self, ulist_index, olist_index, block_index):
        self.ulist_index = ulist_index
        self.olist_index = olist_index
        self.block_index = block_index

    def __str__(self):
        return f"{{ContainerIndices:ulist_index:{self.ulist_index};olist_index:{self.olist_index};block_index:{self.block_index}}}"


# pylint: enable=too-few-public-methods



# pylint: disable=too-many-lines
"""
https://github.github.com/gfm/#link-reference-definitions
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_link_reference_definitions_161():
    """
    Test case 161:  (part 1) A link reference definition does not correspond to a structural element of a document.
    """

    # Arrange
    source_markdown = """[foo]: /url "title"

[foo]"""
    expected_tokens = [
        '[link-ref-def(1,1):True::foo:: :/url:: :title:"title":]',
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/url:title::::foo:::::]",
        "[text(3,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/url" title="title">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_162():
    """
    Test case 162:  (part 2) A link reference definition does not correspond to a structural element of a document.
    """

    # Arrange
    source_markdown = """   [foo]:\a
      /url\a\a
           'the title'\a\a

[foo]""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[link-ref-def(1,4):True:   :foo:: \n      :/url::  \n           :the title:'the title':  ]",
        "[BLANK(4,1):]",
        "[para(5,1):]",
        "[link(5,1):shortcut:/url:the title::::foo:::::]",
        "[text(5,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/url" title="the title">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_163():
    """
    Test case 163:  (part 3) A link reference definition does not correspond to a structural element of a document.
    """

    # Arrange
    source_markdown = """[Foo*bar\\]]:my_(url) 'title (with parens)'

[Foo*bar\\]]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo*bar\\]:Foo*bar\\]::my_(url):: :title (with parens):'title (with parens)':]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:my_(url):title (with parens)::::Foo*bar\\]:::::]",
        "[text(3,2):Foo:]",
        "[text(3,5):*:]",
        "[text(3,6):bar\\\b]:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p><a href="my_(url)" title="title (with parens)">Foo*bar]</a></p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_164():
    """
    Test case 164:  (part 4) A link reference definition does not correspond to a structural element of a document.
    """

    # Arrange
    source_markdown = """[Foo bar]:
<my url>
'title'

[Foo bar]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo bar:Foo bar:\n:my%20url:<my url>:\n:title:'title':]",
        "[BLANK(4,1):]",
        "[para(5,1):]",
        "[link(5,1):shortcut:my%20url:title::::Foo bar:::::]",
        "[text(5,2):Foo bar:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="my%20url" title="title">Foo bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_165():
    """
    Test case 165:  The title may extend over multiple lines:
    """

    # Arrange
    source_markdown = """[foo]: /url '
title
line1
line2
'

[foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: :/url:: :\ntitle\nline1\nline2\n:'\ntitle\nline1\nline2\n':]",
        "[BLANK(6,1):]",
        "[para(7,1):]",
        "[link(7,1):shortcut:/url:\ntitle\nline1\nline2\n::::foo:::::]",
        "[text(7,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/url" title="
title
line1
line2
">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_165a():
    """
    Test case 165a:  The title may extend over multiple lines:
    """

    # Arrange
    source_markdown = """[foo

bar]: /url 'title'

[foo\n\nbar]"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):bar:]",
        "[text(3,4):]:]",
        "[text(3,5):: /url 'title':]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[para(5,1):]",
        "[text(5,1):[:]",
        "[text(5,2):foo:]",
        "[end-para:::True]",
        "[BLANK(6,1):]",
        "[para(7,1):]",
        "[text(7,1):bar:]",
        "[text(7,4):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo</p>
<p>bar]: /url 'title'</p>
<p>[foo</p>
<p>bar]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_166():
    """
    Test case 166:  However, it may not contain a blank line:
    """

    # Arrange
    source_markdown = """[foo]: /url 'title

with blank line'

[foo]"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        "[text(1,6):: /url 'title:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):with blank line':]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[para(5,1):]",
        "[text(5,1):[:]",
        "[text(5,2):foo:]",
        "[text(5,5):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo]: /url 'title</p>
<p>with blank line'</p>
<p>[foo]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_166a():
    """
    Test case 166a:  variation
    """

    # Arrange
    source_markdown = """[foo]: /url 'title
with blank line

[foo]"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        "[text(1,6):: /url 'title\nwith blank line::\n]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[para(4,1):]",
        "[text(4,1):[:]",
        "[text(4,2):foo:]",
        "[text(4,5):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo]: /url 'title\nwith blank line</p>\n<p>[foo]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_166b():
    """
    Test case 166b:  variation
    """

    # Arrange
    source_markdown = """[foo]: /url
'title
with blank line

[foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: :/url:::::]",
        "[para(2,1):\n]",
        "[text(2,1):'title\nwith blank line::\n]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[para(5,1):]",
        "[link(5,1):shortcut:/url:::::foo:::::]",
        "[text(5,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>'title\nwith blank line</p>\n<p><a href="/url">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_167():
    """
    Test case 167:  The title may be omitted:
    """

    # Arrange
    source_markdown = """[foo]:
/url

[foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo::\n:/url:::::]",
        "[BLANK(3,1):]",
        "[para(4,1):]",
        "[link(4,1):shortcut:/url:::::foo:::::]",
        "[text(4,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/url">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_168():
    """
    Test case 168:  The link destination may not be omitted:
    """

    # Arrange
    source_markdown = """[foo]:

[foo]"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        "[text(1,6):::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):[:]",
        "[text(3,2):foo:]",
        "[text(3,5):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo]:</p>
<p>[foo]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_169():
    """
    Test case 169:  However, an empty link destination may be specified using angle brackets:
    """

    # Arrange
    source_markdown = """[foo]: <>

[foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: ::<>::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut::::::foo:::::]",
        "[text(3,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_170():
    """
    Test case 170:  The title must be separated from the link destination by whitespace:
    """

    # Arrange
    source_markdown = """[foo]: <bar>(baz)

[foo]"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        "[text(1,6):: :]",
        "[raw-html(1,8):bar]",
        "[text(1,13):(baz):]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):[:]",
        "[text(3,2):foo:]",
        "[text(3,5):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo]: <bar>(baz)</p>
<p>[foo]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_171():
    """
    Test case 171:  Both title and destination can contain backslash escapes and literal backslashes:
    """

    # Arrange
    source_markdown = """[foo]: /url\\bar\\*baz "foo\\"bar\\baz"

[foo]"""
    expected_tokens = [
        '[link-ref-def(1,1):True::foo:: :/url%5Cbar*baz:/url\\bar\\*baz: :foo&quot;bar\\baz:"foo\\"bar\\baz":]',
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/url%5Cbar*baz:foo&quot;bar\\baz::::foo:::::]",
        "[text(3,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p><a href="/url%5Cbar*baz" title="foo&quot;bar\\baz">foo</a></p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_172():
    """
    Test case 172:  A link can come before its corresponding definition:
    """

    # Arrange
    source_markdown = """[foo]

[foo]: url"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):shortcut:url:::::foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::foo:: :url:::::]",
    ]
    expected_gfm = """<p><a href="url">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_173():
    """
    Test case 173:  If there are several matching definitions, the first one takes precedence:
    """

    # Arrange
    source_markdown = """[foo]

[foo]: first
[foo]: second"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):shortcut:first:::::foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[link-ref-def(3,1):True::foo:: :first:::::]",
        "[link-ref-def(4,1):False::foo:: :second:::::]",
    ]
    expected_gfm = """<p><a href="first">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_174():
    """
    Test case 174:  (part 1) As noted in the section on Links, matching of labels is case-insensitive (see matches).
    """

    # Arrange
    source_markdown = """[FOO]: /url

[Foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:FOO: :/url:::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/url:::::Foo:::::]",
        "[text(3,2):Foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/url">Foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_175():
    """
    Test case 175:  (part 2) As noted in the section on Links, matching of labels is case-insensitive (see matches).
    """

    # Arrange
    source_markdown = """[ΑΓΩ]: /φου

[αγω]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::αγω:ΑΓΩ: :/%CF%86%CE%BF%CF%85:/φου::::]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/%CF%86%CE%BF%CF%85:::::αγω:::::]",
        "[text(3,2):αγω:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/%CF%86%CE%BF%CF%85">αγω</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_176():
    """
    Test case 176:  Here is a link reference definition with no corresponding link. It contributes nothing to the document.
    """

    # Arrange
    source_markdown = """[foo]: /url"""
    expected_tokens = ["[link-ref-def(1,1):True::foo:: :/url:::::]"]
    expected_gfm = """"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_177():
    """
    Test case 177:  Here is another one:
    """

    # Arrange
    source_markdown = """[
foo
]: /url
bar"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:\nfoo\n: :/url:::::]",
        "[para(4,1):]",
        "[text(4,1):bar:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>bar</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_178():
    """
    Test case 178:  This is not a link reference definition, because there are non-whitespace characters after the title:
    """

    # Arrange
    source_markdown = """[foo]: /url "title" ok"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        '[text(1,6):: /url \a"\a&quot;\atitle\a"\a&quot;\a ok:]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo]: /url &quot;title&quot; ok</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_179():
    """
    Test case 179:  This is a link reference definition, but it has no title:
    """

    # Arrange
    source_markdown = """[foo]: /url
"title" ok"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: :/url:::::]",
        "[para(2,1):]",
        '[text(2,1):\a"\a&quot;\atitle\a"\a&quot;\a ok:]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>&quot;title&quot; ok</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_180():
    """
    Test case 180:  This is not a link reference definition, because it is indented four spaces:
    """

    # Arrange
    source_markdown = """    [foo]: /url "title"

[foo]"""
    expected_tokens = [
        "[icode-block(1,5):    :]",
        '[text(1,5):[foo]: /url \a"\a&quot;\atitle\a"\a&quot;\a:]',
        "[end-icode-block:::False]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):[:]",
        "[text(3,2):foo:]",
        "[text(3,5):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<pre><code>[foo]: /url &quot;title&quot;
</code></pre>
<p>[foo]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_181():
    """
    Test case 181:  This is not a link reference definition, because it occurs inside a code block:
    """

    # Arrange
    source_markdown = """```
[foo]: /url
```

[foo]"""
    expected_tokens = [
        "[fcode-block(1,1):`:3::::::]",
        "[text(2,1):[foo]: /url:]",
        "[end-fcode-block::3:False]",
        "[BLANK(4,1):]",
        "[para(5,1):]",
        "[text(5,1):[:]",
        "[text(5,2):foo:]",
        "[text(5,5):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<pre><code>[foo]: /url
</code></pre>
<p>[foo]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_182():
    """
    Test case 182:  A link reference definition cannot interrupt a paragraph.
    """

    # Arrange
    source_markdown = """Foo
[bar]: /baz

[bar]"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):Foo\n::\n]",
        "[text(2,1):[:]",
        "[text(2,2):bar:]",
        "[text(2,5):]:]",
        "[text(2,6):: /baz:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[para(4,1):]",
        "[text(4,1):[:]",
        "[text(4,2):bar:]",
        "[text(4,5):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>Foo
[bar]: /baz</p>
<p>[bar]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183():
    """
    Test case 183:  (part 1) However, it can directly follow other block elements, such as headings and thematic breaks, and it need not be followed by a blank line.
    """

    # Arrange
    source_markdown = """# [Foo]
[foo]: /url
> bar"""
    expected_tokens = [
        "[atx(1,1):1:0:]",
        "[text(1,3)::\a \a\x03\a]",
        "[link(1,3):shortcut:/url:::::Foo:::::]",
        "[text(1,4):Foo:]",
        "[end-link::]",
        "[end-atx::]",
        "[link-ref-def(2,1):True::foo:: :/url:::::]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[text(3,3):bar:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<h1><a href="/url">Foo</a></h1>
<blockquote>
<p>bar</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183a():
    """
    Test case 183a:  variation
    """

    # Arrange
    source_markdown = """---
[foo]: /url
> [Foo]"""
    expected_tokens = [
        "[tbreak(1,1):-::---]",
        "[link-ref-def(2,1):True::foo:: :/url:::::]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[link(3,3):shortcut:/url:::::Foo:::::]",
        "[text(3,4):Foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<hr />
<blockquote>
<p><a href="/url">Foo</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183b():
    """
    Test case 183b:  variation
    """

    # Arrange
    source_markdown = """abc
---
[foo]: /url
> [Foo]"""
    expected_tokens = [
        "[setext(2,1):-:3::(1,1)]",
        "[text(1,1):abc:]",
        "[end-setext::]",
        "[link-ref-def(3,1):True::foo:: :/url:::::]",
        "[block-quote(4,1)::> ]",
        "[para(4,3):]",
        "[link(4,3):shortcut:/url:::::Foo:::::]",
        "[text(4,4):Foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<h2>abc</h2>
<blockquote>
<p><a href="/url">Foo</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183c():
    """
    Test case 183c:  variation
    """

    # Arrange
    source_markdown = """    indented code block
[foo]: /url
> [Foo]"""
    expected_tokens = [
        "[icode-block(1,5):    :]",
        "[text(1,5):indented code block:]",
        "[end-icode-block:::False]",
        "[link-ref-def(2,1):True::foo:: :/url:::::]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[link(3,3):shortcut:/url:::::Foo:::::]",
        "[text(3,4):Foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<pre><code>indented code block
</code></pre>
<blockquote>
<p><a href="/url">Foo</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183d():
    """
    Test case 183d:  variation
    """

    # Arrange
    source_markdown = """```text
indented code block
```
[foo]: /url
> [Foo]"""
    expected_tokens = [
        "[fcode-block(1,1):`:3:text:::::]",
        "[text(2,1):indented code block:]",
        "[end-fcode-block::3:False]",
        "[link-ref-def(4,1):True::foo:: :/url:::::]",
        "[block-quote(5,1)::> ]",
        "[para(5,3):]",
        "[link(5,3):shortcut:/url:::::Foo:::::]",
        "[text(5,4):Foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<pre><code class="language-text">indented code block
</code></pre>
<blockquote>
<p><a href="/url">Foo</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183e():
    """
    Test case 183d:  variation
    """

    # Arrange
    source_markdown = """<script>
<~-- javascript comment -->
</script>
[foo]: /url
> [Foo]"""
    expected_tokens = [
        "[html-block(1,1)]",
        "[text(1,1):<script>\n<~-- javascript comment -->\n</script>:]",
        "[end-html-block:::False]",
        "[link-ref-def(4,1):True::foo:: :/url:::::]",
        "[block-quote(5,1)::> ]",
        "[para(5,3):]",
        "[link(5,3):shortcut:/url:::::Foo:::::]",
        "[text(5,4):Foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<script>
<~-- javascript comment -->
</script>
<blockquote>
<p><a href="/url">Foo</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183f():
    """
    Test case 183e:  variation
    """

    # Arrange
    source_markdown = """> A simple block quote
[foo]: /url
> [Foo]"""
    expected_tokens = [
        "[block-quote(1,1)::> \n\n> ]",
        "[para(1,3):\n\n]",
        "[text(1,3):A simple block quote\n::\n]",
        "[text(2,1):[:]",
        "[text(2,2):foo:]",
        "[text(2,5):]:]",
        "[text(2,6):: /url\n::\n]",
        "[text(3,3):[:]",
        "[text(3,4):Foo:]",
        "[text(3,7):]:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>A simple block quote
[foo]: /url
[Foo]</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, disable_consistency_checks=True
    )


@pytest.mark.gfm
def test_link_reference_definitions_183g():
    """
    Test case 183gx:  variation
    """

    # Arrange
    source_markdown = """- A simple list
[foo]: /url
> [Foo]"""
    expected_tokens = [
        "[ulist(1,1):-::2::]",
        "[para(1,3):\n]",
        "[text(1,3):A simple list\n::\n]",
        "[text(2,1):[:]",
        "[text(2,2):foo:]",
        "[text(2,5):]:]",
        "[text(2,6):: /url:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[text(3,3):[:]",
        "[text(3,4):Foo:]",
        "[text(3,7):]:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<ul>
<li>A simple list
[foo]: /url</li>
</ul>
<blockquote>
<p>[Foo]</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183ga():
    """
    Test case 183gx:  variation
    """

    # Arrange
    source_markdown = """- A simple list
foo]: /url
> [Foo]"""
    expected_tokens = [
        "[ulist(1,1):-::2::]",
        "[para(1,3):\n]",
        "[text(1,3):A simple list\nfoo::\n]",
        "[text(2,4):]:]",
        "[text(2,5):: /url:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[text(3,3):[:]",
        "[text(3,4):Foo:]",
        "[text(3,7):]:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<ul>
<li>A simple list
foo]: /url</li>
</ul>
<blockquote>
<p>[Foo]</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183gb():
    """
    Test case 183gx:  variation
    """

    # Arrange
    source_markdown = """- A simple list
 [foo]: /url
> [Foo]"""
    expected_tokens = [
        "[ulist(1,1):-::2:: ]",
        "[para(1,3):\n]",
        "[text(1,3):A simple list\n::\n]",
        "[text(2,1):[:]",
        "[text(2,2):foo:]",
        "[text(2,5):]:]",
        "[text(2,6):: /url:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[text(3,3):[:]",
        "[text(3,4):Foo:]",
        "[text(3,7):]:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<ul>
<li>A simple list
[foo]: /url</li>
</ul>
<blockquote>
<p>[Foo]</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183gc():
    """
    Test case 183gx:  variation
    """

    # Arrange
    source_markdown = """1. A simple list
 [foo]: /url
> [Foo]"""
    expected_tokens = [
        "[olist(1,1):.:1:3:: ]",
        "[para(1,4):\n]",
        "[text(1,4):A simple list\n::\n]",
        "[text(2,1):[:]",
        "[text(2,2):foo:]",
        "[text(2,5):]:]",
        "[text(2,6):: /url:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[text(3,3):[:]",
        "[text(3,4):Foo:]",
        "[text(3,7):]:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<ol>
<li>A simple list
[foo]: /url</li>
</ol>
<blockquote>
<p>[Foo]</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183gd():
    """
    Test case 183gx:  variation
    """

    # Arrange
    source_markdown = """- A simple list
    [foo]: /url
> [Foo]"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[para(1,3):\n  ]",
        "[text(1,3):A simple list\n::\n]",
        "[text(2,3):[:]",
        "[text(2,4):foo:]",
        "[text(2,7):]:]",
        "[text(2,8):: /url:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[text(3,3):[:]",
        "[text(3,4):Foo:]",
        "[text(3,7):]:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<ul>
<li>A simple list
[foo]: /url</li>
</ul>
<blockquote>
<p>[Foo]</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183ge():
    """
    Test case 183gx:  variation
    """

    # Arrange
    source_markdown = """- A simple list
*foo*: /url
> [Foo]"""
    expected_tokens = [
        "[ulist(1,1):-::2::]",
        "[para(1,3):\n]",
        "[text(1,3):A simple list\n::\n]",
        "[emphasis(2,1):1:*]",
        "[text(2,2):foo:]",
        "[end-emphasis(2,5)::]",
        "[text(2,6):: /url:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[text(3,3):[:]",
        "[text(3,4):Foo:]",
        "[text(3,7):]:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<ul>
<li>A simple list
<em>foo</em>: /url</li>
</ul>
<blockquote>
<p>[Foo]</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183gf():
    """
    Test case 183gx:  variation
    """

    # Arrange
    source_markdown = """- A simple list
items
> [Foo]"""
    expected_tokens = [
        "[ulist(1,1):-::2::]",
        "[para(1,3):\n]",
        "[text(1,3):A simple list\nitems::\n]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[text(3,3):[:]",
        "[text(3,4):Foo:]",
        "[text(3,7):]:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<ul>
<li>A simple list
items</li>
</ul>
<blockquote>
<p>[Foo]</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_183gg():
    """
    Test case 183ga:  variation
    """

    # Arrange
    source_markdown = """- A simple list
[foo]: /url
  > [Foo]"""
    expected_tokens = [
        "[ulist(1,1):-::2::]",
        "[para(1,3):\n]",
        "[text(1,3):A simple list\n::\n]",
        "[text(2,1):[:]",
        "[text(2,2):foo:]",
        "[text(2,5):]:]",
        "[text(2,6):: /url:]",
        "[end-para:::True]",
        "[block-quote(3,3):  :  > ]",
        "[para(3,5):]",
        "[text(3,5):[:]",
        "[text(3,6):Foo:]",
        "[text(3,9):]:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>A simple list
[foo]: /url
<blockquote>
<p>[Foo]</p>
</blockquote>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_184():
    """
    Test case 184:  (part 2) However, it can directly follow other block elements, such as headings and thematic breaks, and it need not be followed by a blank line.
    """

    # Arrange
    source_markdown = """[foo]: /url
bar
===
[foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: :/url:::::]",
        "[setext(3,1):=:3::(2,1)]",
        "[text(2,1):bar:]",
        "[end-setext::]",
        "[para(4,1):]",
        "[link(4,1):shortcut:/url:::::foo:::::]",
        "[text(4,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<h1>bar</h1>
<p><a href="/url">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_185():
    """
    Test case 185:  (part 3) However, it can directly follow other block elements, such as headings and thematic breaks, and it need not be followed by a blank line.
    """

    # Arrange
    source_markdown = """[foo]: /url
===
[foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: :/url:::::]",
        "[para(2,1):\n]",
        "[text(2,1):===\n::\n]",
        "[link(3,1):shortcut:/url:::::foo:::::]",
        "[text(3,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>===
<a href="/url">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_185a():
    """
    Test case 185a:  variation
    """

    # Arrange
    source_markdown = """[foo]: /url
# Abc
[foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: :/url:::::]",
        "[atx(2,1):1:0:]",
        "[text(2,3):Abc: ]",
        "[end-atx::]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/url:::::foo:::::]",
        "[text(3,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<h1>Abc</h1>
<p><a href="/url">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_185b():
    """
    Test case 185b:  variation
    """

    # Arrange
    source_markdown = """[foo]: /url
```text
my text
```
[foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: :/url:::::]",
        "[fcode-block(2,1):`:3:text:::::]",
        "[text(3,1):my text:]",
        "[end-fcode-block::3:False]",
        "[para(5,1):]",
        "[link(5,1):shortcut:/url:::::foo:::::]",
        "[text(5,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<pre><code class="language-text">my text
</code></pre>
<p><a href="/url">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_185c():
    """
    Test case 185c:  variation
    """

    # Arrange
    source_markdown = """[foo]: /url
    my text
[foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: :/url:::::]",
        "[icode-block(2,5):    :]",
        "[text(2,5):my text:]",
        "[end-icode-block:::False]",
        "[para(3,1):]",
        "[link(3,1):shortcut:/url:::::foo:::::]",
        "[text(3,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<pre><code>my text\n</code></pre>\n<p><a href="/url">foo</a></p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_185d():
    """
    Test case 185d:  variation
    """

    # Arrange
    source_markdown = """[foo]: /url
<script>
<~-- javascript comment -->
</script>
[foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: :/url:::::]",
        "[html-block(2,1)]",
        "[text(2,1):<script>\n<~-- javascript comment -->\n</script>:]",
        "[end-html-block:::False]",
        "[para(5,1):]",
        "[link(5,1):shortcut:/url:::::foo:::::]",
        "[text(5,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<script>
<~-- javascript comment -->
</script>
<p><a href="/url">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_185e():
    """
    Test case 185e:  variation
    """

    # Arrange
    source_markdown = """[foo]: /url
> This is a simple block quote
[foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: :/url:::::]",
        "[block-quote(2,1)::> \n]",
        "[para(2,3):\n]",
        "[text(2,3):This is a simple block quote\n::\n]",
        "[link(3,1):shortcut:/url:::::foo:::::]",
        "[text(3,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>This is a simple block quote
<a href="/url">foo</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_185f():
    """
    Test case 185f:  variation
    """

    # Arrange
    source_markdown = """[foo]: /url
- This is a simple list
[foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: :/url:::::]",
        "[ulist(2,1):-::2::]",
        "[para(2,3):\n]",
        "[text(2,3):This is a simple list\n::\n]",
        "[link(3,1):shortcut:/url:::::foo:::::]",
        "[text(3,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>This is a simple list
<a href="/url">foo</a></li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_185fa():
    """
    Test case 185f:  variation
    """

    # Arrange
    source_markdown = """[foo]: /url
- This is a simple list
![foo]"""
    expected_tokens = [
        "[link-ref-def(1,1):True::foo:: :/url:::::]",
        "[ulist(2,1):-::2::]",
        "[para(2,3):\n]",
        "[text(2,3):This is a simple list\n::\n]",
        "[image(3,1):shortcut:/url::foo::::foo:::::]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>This is a simple list
<img src="/url" alt="foo" /></li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_186():
    """
    Test case 186:  Several link reference definitions can occur one after another, without intervening blank lines.
    """

    # Arrange
    source_markdown = """[foo]: /foo-url "foo"
[bar]: /bar-url
  "bar"
[baz]: /baz-url

[foo],
[bar],
[baz]"""
    expected_tokens = [
        '[link-ref-def(1,1):True::foo:: :/foo-url:: :foo:"foo":]',
        '[link-ref-def(2,1):True::bar:: :/bar-url::\n  :bar:"bar":]',
        "[link-ref-def(4,1):True::baz:: :/baz-url:::::]",
        "[BLANK(5,1):]",
        "[para(6,1):\n\n]",
        "[link(6,1):shortcut:/foo-url:foo::::foo:::::]",
        "[text(6,2):foo:]",
        "[end-link::]",
        "[text(6,6):,\n::\n]",
        "[link(7,1):shortcut:/bar-url:bar::::bar:::::]",
        "[text(7,2):bar:]",
        "[end-link::]",
        "[text(7,6):,\n::\n]",
        "[link(8,1):shortcut:/baz-url:::::baz:::::]",
        "[text(8,2):baz:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/foo-url" title="foo">foo</a>,
<a href="/bar-url" title="bar">bar</a>,
<a href="/baz-url">baz</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_187():
    """
    Test case 187:  Link reference definitions can occur inside block containers, like lists and block quotations. They affect the entire document, not just the container in which they are defined:
    """

    # Arrange
    source_markdown = """[foo]

> [foo]: /url"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):shortcut:/url:::::foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[block-quote(3,1)::> ]",
        "[link-ref-def(3,3):True::foo:: :/url:::::]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<p><a href="/url">foo</a></p>
<blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_188():
    """
    Test case 188:  Whether something is a link reference definition is independent of whether the link reference it defines is used in the document.
    """

    # Arrange
    source_markdown = """[foo]: /url"""
    expected_tokens = ["[link-ref-def(1,1):True::foo:: :/url:::::]"]
    expected_gfm = """"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_188a():
    """
    Test case 188a:  variation
    """

    # Arrange
    source_markdown = """["""
    expected_tokens = ["[para(1,1):]", "[text(1,1):[:]", "[end-para:::True]"]
    expected_gfm = """<p>[</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_188b():
    """
    Test case 188b:  variation
    """

    # Arrange
    source_markdown = """[foo"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_188c():
    """
    Test case 188c:  variation
    """

    # Arrange
    source_markdown = """[foo]"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_188d():
    """
    Test case 188d:  variation
    """

    # Arrange
    source_markdown = """[foo]:"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        "[text(1,6):::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo]:</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_188e():
    """
    Test case 188e:  variation
    """

    # Arrange
    source_markdown = """[foo]: /url ("""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[text(1,5):]:]",
        "[text(1,6):: /url (:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo]: /url (</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_extra_01():
    """
    Test case extra 01:  variation
    """

    # Arrange
    source_markdown = """- [foo]:
/url"""
    expected_tokens = [
        "[ulist(1,1):-::2::]",
        "[link-ref-def(1,3):True::foo::\n:/url:::::]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li></li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_extra_01a():
    """
    Test case extra 01:  variation
    """

    # Arrange
    source_markdown = """- [foo]:
  /url"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[link-ref-def(1,3):True::foo::\n:/url:::::]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li></li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_extra_01b():
    """
    Test case extra 01:  variation
    """

    # Arrange
    source_markdown = """- [foo]:
 /url"""
    expected_tokens = [
        "[ulist(1,1):-::2:: ]",
        "[link-ref-def(1,3):True::foo::\n:/url:::::]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li></li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_extra_01c():
    """
    Test case extra 01:  variation
    """

    # Arrange
    source_markdown = """- [foo]:
- /url"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):[:]",
        "[text(1,4):foo:]",
        "[text(1,7):]:]",
        "[text(1,8):::]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):/url:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>[foo]:</li>
<li>/url</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_extra_01d():
    """
    Test case extra 01:  variation
    """

    # Arrange
    source_markdown = """1. [foo]:
1. /url"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):[:]",
        "[text(1,5):foo:]",
        "[text(1,8):]:]",
        "[text(1,9):::]",
        "[end-para:::True]",
        "[li(2,1):3::1]",
        "[para(2,4):]",
        "[text(2,4):/url:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>[foo]:</li>
<li>/url</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_extra_02x():
    """
    Test case extra 02:  variation
    """

    # Arrange
    source_markdown = """> [foo]:
/url"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[link-ref-def(1,3):True::foo::\n:/url:::::]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_extra_02a():
    """
    Test case extra 02a:  variation
    """

    # Arrange
    source_markdown = """> [foo]:
> /url"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[link-ref-def(1,3):True::foo::\n:/url:::::]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_extra_02bx():
    """
    Test case extra 02b:  variation
    """

    # Arrange
    source_markdown = """> [foo]:
>> /url"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[para(1,3):]",
        "[text(1,3):[:]",
        "[text(1,4):foo:]",
        "[text(1,7):]:]",
        "[text(1,8):::]",
        "[end-para:::True]",
        "[block-quote(2,1)::>> ]",
        "[para(2,4):]",
        "[text(2,4):/url:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>[foo]:</p>
<blockquote>
<p>/url</p>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_extra_02ba():
    """
    Test case extra 02b:  variation
    """

    # Arrange
    source_markdown = """> [foo]:
>>> /url"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[para(1,3):]",
        "[text(1,3):[:]",
        "[text(1,4):foo:]",
        "[text(1,7):]:]",
        "[text(1,8):::]",
        "[end-para:::True]",
        "[block-quote(2,1)::]",
        "[block-quote(2,2)::>>> ]",
        "[para(2,5):]",
        "[text(2,5):/url:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>[foo]:</p>
<blockquote>
<blockquote>
<p>/url</p>
</blockquote>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_extra_02cx():
    """
    Test case extra 02c:  variation
    """

    # Arrange
    source_markdown = """>> [foo]:
> /url"""
    expected_tokens = [
        "[block-quote(1,1)::]",
        "[block-quote(1,2)::>> \n> ]",
        "[link-ref-def(1,4):True::foo::\n:/url:::::]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_link_reference_definitions_extra_02ca():
    """
    Test case extra 02c:  variation
    """

    # Arrange
    source_markdown = """>>> [foo]:
> /url"""
    expected_tokens = [
        "[block-quote(1,1)::]",
        "[block-quote(1,2)::]",
        "[block-quote(1,3)::>>> \n> ]",
        "[link-ref-def(1,5):True::foo::\n:/url:::::]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
<blockquote>
</blockquote>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)
